      program PREPARE3D

      Integer*4 NN
      Parameter (NN=5000)


C     Fortran version: Mark Hopkins 9/11/89; revised Michel Louge 1/10/94.
C     Preparation program for 3D Couette flows.
C     This version calculates the x,y & z dimensions of the flow domain and
C     the value of nu near a specified value.
C     The particles are placed in a slightly dilated tetrahedral
C     array.

C     Flat walls are at top and bottom.

C     The shear rate Gamma is hardwired to one for simplicity. MYL 8/9/1990.

C This version is updated to allow both top and bottom bumpy walls.
C It also permits interactive setup of the particle concentration
C and periodic domain dimensions. MYL 6/26/91.

C Modified 10/18/1990 to let the top wall particles behave like a thermostat
C of a specified temperature.

C     Note: The domain is divided into cells whose diagonal is 0.92 of a
C     disk diameter to insure that only one disk occupies a cell at any
C     time.  The length and width of the flow domain must in turn be
C     divisible into some unit number of cells.  Therefore the length and
C     the width of the flow domain must be recalculated and adjusted slightly
C     in terms of the cell width and the disk diameter.  If some exact value 
C     of width to diameter is wanted this can be a bit of a problem.

      Integer*4  i,N,Nu,Nl,Np,ni,nj,nk,Nx,Nz,Npx,Npy,Npz
	integer*4 Nxmax,Nzmax,Nwallx,Nwallz,k
	
      Real*8 x(NN),y(NN),z(NN),u(NN),v(NN),w(NN)
      Real*8 xo,yo,zo,xp,yp,zp,xav,yav,zav,uav,vav,wav
      Real*8 ed,ew,mu,muw,D,cell,pi,xlen,ylen,zlen,nul,vwall
      real*8 beta0,beta0w
      Real*8 con,dt,time,dx,dy,dz,volume,sD,gravx,gravy
      Real*8 std,sumu,sumw,sumT,correct,xmin,xmax,ymin,ymax,zmin,zmax,
     &       dil,dilx,dily,dilz,prepco
      real*8 rand,fluc,aaa,bbb,newylen
      Character*12 fname  
      Integer*4 flagb,flagt,response

********************************ox**********************************

      data nul,xo,yo,zo/4*0.0/
      data xav,yav,zav,uav,vav,wav/6*0.0/

      write (*,'(A)') ' ? Configuration File ?'
      Read (*,'(A)') fname

      pi = rand(13)
      pi = 3.1415926535
      time = 0.0

      write (*,'(A)') ' ? bumpy top wall; yes: 1 ?'
      Read (*,*) flagt
      If (flagt.ne.1) then
      	flagt = 0
	Nu=0
	write (*,'(a)') 'no top boundary particles'
      end if

      write (*,'(A)') ' ? bumpy bottom wall; yes: 1 ?'
      Read (*,*) flagb
      If (flagb.ne.1) then
      	flagb = 0
	Nl=0
	write (*,'(a)') 'no bottom boundary particles'
      end if

	write (*,'(/,a,/)') 'begin interactive session'
	
      write (*,'(A)') ' ? ylen (distance between plates) ?'
      Read (*,*) ylen
      
111	write (*,'(A)')
     & ' ? prepco (cell width=diameter*prepco/sqrt(3)) ?'
      write (*,'(A)')
     &'Note: prepco must be strictly between sqrt(3)/2 & 1'
      Read (*,*) prepco
      if (prepco.le.(sqrt(3.)/2.).or.prepco.ge.1) go to 111

      write (*,'(A)')
     & ' ? ni (number of cells along the domain length) ?'
      Read (*,*) ni

      write (*,'(A)')
     & ' ? nj (number of cells along the domain width) ?'
      Read (*,*) nj

      write (*,'(A)')
     & ' ? nk (number of cells along the domain depth) ?'
      Read (*,*) nk
      
      d = ylen*sqrt(3.)/float(nj)/prepco
      cell = ylen/float(nj)
      xlen = float(ni)*cell
      zlen = float(nk)*cell
      
      newylen = ylen - d/2.*(flagb+flagt)
      volume = xlen*newylen*zlen
      
      write (*,'(A)')
     & ' ? desired particle volume fraction ?'
      Read (*,*) con
      
      N = int((6.0*con*Volume)/(pi*D*D*D))
      
      write (*,'(A,i10)')
     & 'no of particles to achieve volume fraction = ',N

      if (N.ge.nn) then
      write (*,'(a)') 'N is greater than NN'
      stop
      end if
      

      con = N*pi*D*D*D/(6.0*Volume)
      dil = (pi/6.*sqrt(2.)/con)**(1./3.)
      dilx = dil
      dily = dil
      dilz = dil
      
C calculate the number of particles fitting in each direction
C on the basis of a hexagonal close-pack
C given the average dilatation and the desired concentration:
      
      Npx = int( xlen/dilx/d )
      Npy = int( (newylen-d*dily)/dily/sqrt(2./3.)/d ) + 1
      Npz = int( zlen/dilz/d*2./sqrt(3.) )
      
      N = Npx*Npy*Npz
      
      if (N.ge.nn) then
      write (*,'(a)') 'N is greater than NN'
      stop
      end if
      
      con = N*pi*D*D*D/(6.0*Volume)

*** Write the current estimates with the following format:

***dilx   dily   dilz   nk   Npx  Npy  Npz    N     con
***1.253  1.253  1.253  128  128  128  128  20000  30.50

3     write (*,'(/,a,/)') 
     &'dilx   dily   dilz   nk   Npx  Npy  Npz    N     con'
      write (*,'(3(f5.3,2x),4(i3,2x),i5,2x,f5.2,/)') 
     & dilx,dily,dilz,nk,Npx,Npy,Npz,N,100.*con
     
C If the number of rows in the z-direction is odd, then
C particles with the largest value of z may interfere
C with the periodic particles on top at the initial time:

     	if (mod(Npz,2).eq.1) then
	write (*,'(/,a,/,a)')
     &  'Odd number of particle rows in the z-direction:',
     &  'Change dilz or nk'
     	go to 1
	end if
	
	write (*,'(a)') 'Are you satisfied (yes=1) ?'
	
	read (*,*) response
	
		if (response.eq.1) then
		
	write (*,'(3(/,a))')
     &'OK to calculate the following configuration?',
     &' (yes=1, no=0, stop<0) ',
     &'xlen    ylen    zlen    ni   nj   nk     N     con'
	write (*,'(3(f6.3,2x),3(i3,2x),i5,2x,f5.2,/)') 
     & xlen,ylen,zlen,ni,nj,nk,N,100.*con
     
*** Write the current configuration with the following format:

***xlen    ylen    zlen    ni   nj   nk     N     con
***10.253  10.253  10.253  128  128  128  20000  30.50

	read (*,*) response
			if (response.eq.1) then
			
C Calculate the fluffiest configuration possible to avoid wholesale
C overlaps at time zero:

C Intervals between successive particles in the 3 directions:

		dx = xlen/float(Npx)
		dz = zlen/float(Npz)
		dy = newylen/ (float(Npy - 1) + sqrt(3./2.) )
		dily = dy/sqrt(2./3.)/d
		
		go to 2
			else
				if (response.lt.0) then
				stop
				else
				go to 1
				end if
				
			end if
	
		else
	
		go to 1
		
		end if
		
1	continue

10	write (*,'(a)') 'enter new value of dilx (>1)'
	read (*,*) dilx
	if (dilx.lt.1) go to 10
	
11	write (*,'(a)') 'enter new value of dily (>1)'
	read (*,*) dily
	if (dily.lt.1) go to 11
	
12	write (*,'(a)') 'enter new value of dilz (>1)'
	read (*,*) dilz
	if (dilz.lt.1) go to 12
	
	write (*,'(a)') 'enter new value of nk'
	read (*,*) nk
	
C New estimates:

      zlen = float(nk)*cell
      Npx = int( xlen/dilx/d )
      Npy = int( (newylen-d*dily)/dily/sqrt(2./3.)/d ) + 1
      Npz = int( zlen/dilz/d*2./sqrt(3.) )
      N = Npx*Npy*Npz
      
      if (N.ge.nn) then
      write (*,'(a)') 'N is greater than NN'
      stop
      end if
      
      volume = xlen*newylen*zlen
      con = N*pi*D*D*D/(6.0*Volume)
      
      go to 3
	
2	continue

      write (*,'(A,/,a)') ' ? Initial time step ?',
     &      'recommended: (d/10)(2/Gamma/ylen)(Pi/6/nu)^1/3 '
      Read (*,*) dt
      write (*,'(A)') ' ? coef of restitution (interior) ?'
      Read (*,*) ed
      write (*,'(A)') ' ? coef of restitution (at walls) ?'
      Read (*,*) ew
      write (*,'(A)')
     & ' ? coef of friction (interior & wall particles) ?'
      Read (*,*) mu
      write (*,'(A)') ' ? coef of friction at the walls?'
      Read (*,*) muw
      write (*,'(A)')
     & ' ? coef of tang rest (int & wall particles) ?'
      Read (*,*) beta0
      write (*,'(A)') ' ? coef of tang rest at the walls?'
      Read (*,*) beta0w

CCCC
C The wall velocity is such that the shear rate Gamma is one.
C The top wall moves at VWALL and the bottom moves at -VWALL:
CCCC

      Vwall = ylen/2.
      
CCC
C The maximum number of wall spheres in either directions is
C calculated by assuming that the spheres pack on a square 
C array. A larger no of spheres could be fitted by assuming
C hexagonal packing. Note that the diameter of each wall sphere
C is equal to that of the interior particles.
CCC
		if (flagt.ne.0.or.flagb.ne.0) then

	Nxmax = int(xlen/d)
	Nzmax = int(zlen/d)
	
	write (*,'(a,/,a,I3)')
     &  'no of wall spheres in the x-direction?',
     &  'maximum is ',Nxmax
     
     	read (*,*) Nwallx
	
	write (*,'(a,/,a,I3)')
     &  'no of wall spheres in the z-direction?',
     &  'maximum is ',Nzmax

     	read (*,*) Nwallz
	
	if (Nwallx.gt.Nxmax) then
	
	Nwallx = Nxmax
	write (*,'(a,I3)')
     &  'Adjusted no of spheres in the x-direction: ',Nwallx
     
     	end if
	
	if (Nwallz.gt.Nzmax) then
	
	Nwallz = Nzmax
	write (*,'(a,I3)')
     &  'Adjusted no of spheres in the z-direction: ',Nwallz
     
     	end if
	
CCC
C Calculate the no of spheres:
CCC

	if (flagb.eq.1) Nl = Nwallx*Nwallz
	if (flagt.eq.1) Nu = Nwallx*Nwallz
	
	if (nu.ne.0) then

	write (*,'(A)')
     &  ' ? 2-D granular temperature of the top wall particles?'
      Read (*,*) std
      
      std = sqrt(2.*std)
      
      	end if
		end if

C sD is a fossil variable left over from the disk simulations.

      sD = 0.0
C
CCC The following may be changed to have gravity in the x-direction
C (gravx) or the y-direction (gravy):

      gravx = 0.0
      gravy = 0.0

      Open(1, File=fname, Form='Formatted', Status='New')

      write (1,'(3I15)') N,Nl,Nu
      write (1,'(3I15)') ni,nj,nk
      write (1,'(2F15.10)') time,dt
      write (1,'(6F10.5)') ed,ew,mu,muw,beta0,beta0w
      write (1,'(3F15.10)') 0.5*D,cell,con
      write (1,'(3F15.10)') sD,gravx,gravy
      write (1,'(3F15.10)') xlen,ylen,zlen

C Origin of coordinate system:

      xo = 0.0
      yo = 0.0
      zo = 0.0

C Position of the first interior particle is the following + dx:

      xp = xo
      yp = d/2.*flagb + d*dily/2. + yo
      zp = zo
      
C Initialize counters of particles in the x-direction (Nx),
C of particle rows in the z-direction (Nz), and of particle
C layers in the y-direction (Ny):

      Nx = 0
      Ny = 1
      Nz = 1
      Np = 0

C     Fill x, then z, then y.

C Check maximum distance in each direction:

	xmax = 0.0
	xmin = xlen

	ymax = 0.0
	ymin = ylen

	zmax = 0.0
	zmin = zlen

      Do While (Np .lt. N)

        Np = Np+1
        Nx = Nx+1

        xp = xp+dx
        x(Np) = xp
        y(Np) = yp
        z(Np) = zp

        xav = xav+x(Np)
        yav = yav+y(Np)
        zav = zav+z(Np)

        if (mod(Np,2) .eq. 1) then
	
          v(Np) = sign(Vwall*std,0.5-rand(Np))
	  
C
C To start, it is actually advantageous not to have any velocity
C component in the x- or the z-directions:

C          w(Np) = sign(Vwall*std,0.5-rand(Np+0.1))

        else
	
          v(Np) = -v(Np-1)
	  
C         w(Np) = -w(Np-1)

        end if

	u(Np) = 0.0
	w(Np) = 0.0
	
C If the row of particles is complete in the x-direction,
C reset the counters, move to the next row, and calculate
C the position of the first particle in the next row:

        if (Nx .eq. Npx) then
          Nx = 0
          Nz = Nz+1

C Note: particles are staggered in two consecutive rows
C according to the hexagonal close-pack arrangement:

          if (mod(Ny,2) .eq. 1) then
            xp = xo - 0.5*mod(Nz-1,2)*dx        ! odd layers in y
          else
            xp = xo - 0.5*mod(Nz,2)*dx          ! even layers in y
          end if

          zp = zp+dz
        end if 

C If all rows are complete in a given layer, then reset also the
C Nz counter, and recalculate the position of the first particle
C in the first row of the next y-layer:

        if (Nz .gt. Npz) then
          Ny = Ny+1
          Nz = 1

          yp = yp+dy

C Note: particles are staggered in two consecutive layers
C according to the hexagonal close-pack arrangement:

          if (mod(Ny,2) .eq. 1) then
            xp = xo                    ! odd layers
          else               
            xp = xo - 0.5*dx             ! even layers
          end if

          zp = zo+mod(Ny-1,2)*dz/3.0
        end if 

C The following instruction ensures that if the number of
C interior particles is not even, the average velocity will
C nevertheless be zero by cancelling the velocity of the last
C particle:

      if (Np.eq.N.and.mod(N,2) .eq. 1) then

        v(N) = 0.0

      end if

C Write the position and velocity of the particles:

        write (1,'(4F15.10)') x(Np),y(Np),z(Np),0.5*D
        write (1,'(2(e19.13,2x),e19.13)') u(Np),v(Np),w(Np)
        write (1,'(2(e19.13,2x),e19.13)') nul,nul,nul

        uav = uav+u(Np)
        vav = vav+v(Np)
        wav = wav+w(Np)

C find the extremum locations in each direction

		if (x(Np).le.xmin) xmin = x(Np)
		if (x(Np).ge.xmax) xmax = x(Np)

		if (y(Np).le.ymin) ymin = y(Np)
		if (y(Np).ge.ymax) ymax = y(Np)

		if (z(Np).le.zmin) zmin = z(Np)
		if (z(Np).ge.zmax) zmax = z(Np)

C Leave the program if particles are outside the central periodic
C domain:

		if (x(Np).lt.0.0.or.x(Np).gt.xlen) then
			close (1)
			write (*,'(a)') 'particle out in the x-direction'
			stop
		end if

		if (y(Np).gt. (ylen-d/2.*(flagt+1.)) ) then
			close (1)
			write (*,'(a)')
     &			'particle interferes with the top boundary'
			stop
		end if
		
		if (y(Np).lt. (d/2.*(flagb+1.)) ) then
			close (1)
			write (*,'(a)')
     &			'particle interferes with the bottom boundary'
			stop
		end if
		
		if (z(Np).lt.0.0.or.z(Np).gt.zlen) then
			close (1)
			write (*,'(a)') 'particle out in the z-direction'
			stop
		end if
		
      End Do

      xav = xav/N
      yav = yav/N
      zav = zav/N

C Write the location of the center of gravity, the total momentum and
C the extremum positions of the particles in each direction:

      	write (*,'(a,/,6(a,f8.3,/),/)')
     &  'Extremum positions of the particles: ',
     &	' xmin = ',xmin,
     &	' xmax = ',xmax,
     &	' ymin = ',ymin,
     &	' ymax = ',ymax,
     &	' zmin = ',zmin,
     &	' zmax = ',zmax
     
      	write (*,'(a,/,3(a,f8.3,/),/)')
     &  'Position of the center of gravity: ',
     &	' xav = ',xav,
     &	' yav = ',yav,
     &	' zav = ',zav

      	write (*,'(a,/,3(a,f8.3,/),/)')
     &  'Average momentum in 3 directions: ',
     &	' uav = ',uav,
     &	' vav = ',vav,
     &	' wav = ',wav

CCC
C Insert wall spheres after the interior particles and
C before the flat walls (lower wall first).
C Their average velocity is that of the
C wall and they do not rotate. In this program, their radius
C is identical to that of the interior particles:
CCC

	If (nl.ne.0) then
	
	np = n

	do i=1,Nwallx
	do k=1,Nwallz
	
	np = np + 1
	
	write (1,'(4F15.10)')
     &  (i-1.)*xlen/float(Nwallx),nul,
     &  (k-1.)*zlen/float(Nwallz),0.5*D
	
	u(np) = - vwall
	w(np) =   0.0

	write (1,'(2(e19.13,2x),e19.13)') u(np) , nul , w(np)

	write (1,'(2(e19.13,2x),e19.13)') nul,nul,nul

	end do
	end do
	
	end if
	
C Upper wall has a granular temperature.
C  First, calculate positions and uncorrected random velocities:

	If (nu.ne.0) then
	
	np = n+nl
	sumu = 0.0
	sumw = 0.0
	sumT = 0.0
	
	do i=1,Nwallx
	do k=1,Nwallz
	
	np = np + 1
	
	aaa=rand(np)
	bbb=fluc(aaa)

		if (rand(np+0.2).ge.0.5) then
	u(np) = + std * bbb
		else
	u(np) = - std * bbb
		end if
		
	
C	write (*,'(3(a,f10.7))')
C    &  'aaa = ',aaa,' bbb = ',bbb,' u = ',u(np)
	
	aaa=rand(np+0.1)
	bbb=fluc(aaa)
	
		if (rand(np+0.3).ge.0.5) then
	w(np) = + std * bbb
		else
	w(np) = - std * bbb
		end if
		
C	write (*,'(3(a,f10.7))')
C    &  'aaa = ',aaa,' bbb = ',bbb,' w = ',w(np)
	
	sumu = sumu + u(np)
	sumw = sumw + w(np)
	sumT = sumT + u(np)*u(np) + w(np)*w(np)

	end do
	end do
	
C Rescale the fluctuating velocities with the factor CORRECT:

		if (std.ne.0.0) then
	correct = std/sqrt(sumT/nu-(sumu/nu)**2-(sumw/nu)**2)
		else
	correct = 1.0
		end if
	
C Write the positions of the wall particles and their velocities:

	np = n+nl

	do i=1,Nwallx
	do k=1,Nwallz
	
	np = np + 1
	
	write (1,'(4F15.10)')
     &  (i-1.)*xlen/float(Nwallx),ylen,
     &  (k-1.)*zlen/float(Nwallz),0.5*D
	
C Correct the velocities of the boundary particles to ensure that
C they move at the wall velocity and that their temperature is
C equal to the specified temperature:

	u(np) = (u(np) - sumu/float(nu))*correct + vwall
	w(np) = (w(np) - sumw/float(nu))*correct

	write (1,'(2(e19.13,2x),e19.13)') u(np) , nul , w(np)

	write (1,'(2(e19.13,2x),e19.13)') nul,nul,nul

	end do
	end do
	
	end if
	
C     Flat walls: N+1 = bottom, N+2 = top.

      write (1,'(4F15.10)') 0.5*xlen,nul,0.5*zlen,nul
      write (1,'(2(e19.13,2x),e19.13)') -vwall,nul,nul
      write (1,'(2(e19.13,2x),e19.13)') nul,nul,nul

      write (1,'(4F15.10)') 0.5*xlen,ylen,0.5*zlen,nul
      write (1,'(2(e19.13,2x),e19.13)') vwall,nul,nul
      write (1,'(2(e19.13,2x),e19.13)') nul,nul,nul
 
C Check the average and fluctuating velocities of the wall particles:

	If (nl.ne.0) then
	
	np = n
	sumu = 0.0
	sumw = 0.0
	sumT = 0.0
	
	do i=1,Nwallx
	do k=1,Nwallz
	
	np = np + 1
	
	sumT = sumT + u(np)*u(np) + w(np)*w(np)
	sumu = sumu + u(np)
	sumw = sumw + w(np)
	
	end do
	end do
	
	write (*,'(a)') ' Bottom boundary particles: '
	write (*,'(2(a,f13.6))') ' v1 = ',sumu/nl,' v3 = ',sumw/nl
	write (*,'(a,e12.5)') 'actual 2-D temperature = ',
     &  (sumT/nl - (sumu/nl)**2 - (sumw/nl)**2)/2.
	
	end if
	
	If (nu.ne.0) then
	
	np = n+nl
	sumu = 0.0
	sumw = 0.0
	sumT = 0.0
	
	do i=1,Nwallx
	do k=1,Nwallz
	
	np = np + 1
	
	sumT = sumT + u(np)*u(np) + w(np)*w(np)
	sumu = sumu + u(np)
	sumw = sumw + w(np)
	
	end do
	end do
	
	write (*,'(a)') ' Top boundary particles: '
	write (*,'(2(a,f13.6))') ' v1 = ',sumu/nu,' v3 = ',sumw/nu
	write (*,'(a,e12.5)') 'actual 2-D temperature = ',
     &  (sumT/nu - (sumu/nu)**2 - (sumw/nu)**2)/2.
	
	end if
	
	stop
	
      End

********************************************************************

	function fluc (x)
	
	real*8 x,fluc
	
CCCC
C This function calculates an approximate value of the 
C reciprocal of the error function. If x is a random number
C equiprobable between 0 and 1, then the fluctuating
C velocity components u = sqrt(2 T) erf-1(x) in each three directions
C are distributed like a gaussian with granular temperature T.
CCCC
	
	fluc = (0.44035*x + 153.513*x**2 - 2124.81*x**3 +
     &         14756.7*x**4 - 58661.6*x**5 + 141889.*x**6 -
     &         212418.*x**7 + 192214.*x**8 - 96322.1*x**9 +
     &         20522.9*x**10)/5.
     
     	return
     	end
	
